@@ -1,6 +1,6 @@
 /*
  * GTK VNC Widget
- *  
+ *
  * Copyright (C) 2006  Anthony Liguori <anthony@codemonkey.ws>
  *
  * This library is free software; you can redistribute it and/or
@@ -137,6 +137,7 @@ void		vnc_display_force_grab(VncDisplay *obj, gboolean enable);
 gboolean	vnc_display_is_pointer_absolute(VncDisplay *obj);
 
 GOptionGroup *  vnc_display_get_option_group(void);
+const GOptionEntry *  vnc_display_get_option_entries(void);
 
 G_END_DECLS
 
